/*

(C) 2009 Mika Ilmaranta <ilmis@nullnet.fi>

License: GPLv2

*/

#include <stdio.h>
#include <errno.h>
#include <time.h>
#include <signal.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stddef.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <netinet/icmp6.h>
#include <netinet/in_systm.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/select.h>
#include <syslog.h>
#include <sys/wait.h>
#include <stdarg.h>
#include <fcntl.h>

#include <sys/ioctl.h>
#include <linux/if.h>
#include <linux/if_arp.h>
#include <sys/uio.h>

#include "icmp_t.h"
#include "config.h"

#ifndef TRUE
#define TRUE	1
#define FALSE	0
#endif

#define TIMEVAL_DIFF_CMP_GT 0
#define TIMEVAL_DIFF_CMP_LT 1

#define MIN_PERHOST_INTERVAL 20000L /* 20ms in between sends minimum */
#define DEFAULT_SELECT_WAIT  10000L /* wait at least 10ms for incoming packet */

#define FOLLOWED_PKTS 100 /* THIS ABSOLUTELY CAN'T EXCEED 0x10000 (65536 decimal) OR THINGS BREAK */
#define SEQ_LIMITER   ((0x10000 / FOLLOWED_PKTS) * FOLLOWED_PKTS)

typedef struct sentpkt {
	unsigned short seq;
	struct timeval sent_time;
	struct timeval replied_time;
	unsigned long rtt;
	struct {
		unsigned replied:1;
		unsigned timeout:1;
		unsigned waiting:1;
		unsigned used:1;
		unsigned error:1;
	} flags;
} SENTPKT;

typedef struct target {
	unsigned short id; /* target id */
	unsigned short seq;
	struct sockaddr_in src_addr;
	struct sockaddr_in dst_addr;
	struct sockaddr_ll me;
	struct sockaddr_ll he;
	struct in_addr src;
	struct in_addr dst;
	unsigned long num_sent;
	struct timeval last_send_time;
	int status;
	int sock;
	SENTPKT sentpkts[FOLLOWED_PKTS];
} TARGET;

typedef struct ping_data {
	unsigned short id;       /* target id */
	long ping_count;         /* counts up to -c count or 1 */
	struct timeval ping_ts;  /* time sent */
} PING_DATA;

typedef struct status_data {
	char *name;
	int status;
	struct status_data *next;
} STATUS_DATA;

void decide(CONFIG *first);
void groups_decide(GROUPS *firstg);
int wait_for_replies(CONFIG **ctable);
int ping_send(CONFIG *cur);
int ping_rcv(CONFIG *first, char *buf, int len, struct sockaddr_in *saddr, unsigned int *slen, long usec, CONFIG **arp);
int handle_odd_icmp(struct icmp *icp, int icpsize, struct sockaddr_in *saddr);
int in_cksum(u_short *p, int n);
int timeval_diff_cmp(struct timeval *a, struct timeval *b, int operation, time_t sec, suseconds_t usec);
long timeval_diff(struct timeval *a, struct timeval *b);
void usage(void);
void forkexec(char *fmt, ...);
int event_script_check(const char *path);
void signal_handler(int signo);
void makeplist(char *plist[], char *buf);
static int reopen(CONFIG *cur);
static void init_config(CONFIG *first, CONFIG *last, struct protoent *proto, CONFIG ***ctable);
static void free_config_data(CONFIG *first);

#define min(x,y) ((x)<(y) ? (x) : (y))

static int cont = TRUE;
static int dump = FALSE;
static int ident = 0;
static int reload_cfg = FALSE;

char *prog;

/* Main */
int main(int argc, char *argv[]) {
	struct protoent *proto;
	TARGET *t = NULL;
	CONFIG *first = NULL, *last = NULL, *cur;
	GROUPS *firstg = NULL, *lastg = NULL;
	CONFIG **ctable = NULL;
	struct timeval last_sent_time = {0, 0};
	int start = 0;
	int lfp = 0;

	prog = argv[0];
	ident = getpid() & 0xFFFF;

	openlog("lsm", LOG_PID, LOG_DAEMON);

	if(read_config(argv[1], &first, &last, &firstg, &lastg)) {
		usage();
		exit(2);
	}

	if(cfg.debug >= 9) syslog(LOG_INFO, "my ident is %d\n", ident);

	if(!first) {
		syslog(LOG_ERR, "no targets found in config file");
		exit(1);
	}

	if(cfg.debug >= 9) dump_config(&first, &last, &firstg, &lastg);

	/* check pid file */
	if(argc >= 3) {
		lfp = open(argv[2], O_RDWR|O_CREAT, 0640);

		if(lfp < 0) {
			syslog(LOG_ERR, "can't open pid file %s", argv[2]);
			exit(1); /* can not open */
		}
		if(fcntl(lfp, F_SETFD, FD_CLOEXEC) == -1) {
			syslog(LOG_ERR, "failed to set close on exec on pid file %s", argv[2]);
		}

		if(lockf(lfp, F_TLOCK, 0) < 0) {
			syslog(LOG_ERR, "can't lock pid file %s", argv[2]);
			exit(0); /* can not lock */
		}
		/* only first instance continues */
	}

	/* detach from controlling terminal if debug level is below 100 */
	if(cfg.debug < 100) {
		if(daemon(1, 0)) {
			syslog(LOG_ERR, "daemon failed while trying to detach");
			return(1);
		}
	}

	if(lfp) {
		char str[BUFSIZ];
		ssize_t n;

		lseek(lfp, 0, SEEK_SET);
		if(ftruncate(lfp, 0) == -1) {
			syslog(LOG_ERR, "ftruncate failed \"%s\"", strerror(errno));
			exit(1);
		}

		sprintf(str, "%d\n", getpid());
		n = write(lfp, str, strlen(str)); /* record pid to lockfile */
		if(n == -1) {
			syslog(LOG_ERR, "write failed \"%s\"", strerror(errno));
			exit(1);
		}
		if(n != strlen(str)) {
#if defined(__x86_64) || defined(__x86_64__) || defined(__amd64) || defined(__amd64__)
			syslog(LOG_ERR, "write failed, %ld bytes written of %ld bytes", n, strlen(str));
#else
			syslog(LOG_ERR, "write failed, %d bytes written of %d bytes", n, strlen(str));
#endif
			exit(1);
		}
	}

	if((proto = getprotobyname("icmp")) == NULL) {
		syslog(LOG_ERR, "no icmp proto found");
		exit(1);
	}

	init_config(first, last, proto, &ctable);

	signal(SIGINT, signal_handler);
	signal(SIGUSR1, signal_handler);
	signal(SIGHUP, signal_handler);

	/* the main loop */
	while(cont) {
		struct timeval tv = {0, 0};
		struct timeval last_decision = {0, 0};

		if(reload_cfg) {
			STATUS_DATA *first_status = NULL;
			STATUS_DATA *last_status = NULL;
			STATUS_DATA *cur_status = NULL;

			reload_cfg = 0;
			/* store connection statuses temporarily */
			for(cur = first; cur; cur = cur->next) {
				if((cur_status = malloc(sizeof(STATUS_DATA))) == NULL) {
					syslog(LOG_ERR, "could not malloc for status_data");
					exit(1);
				}
				cur_status->name = strdup(cur->name);
				cur_status->status = ((TARGET *)cur->data)->status;

				if(last_status) {
					last_status->next = cur_status;
					last_status = cur_status;
					cur_status->next = NULL;
				} else {
					first_status = cur_status;
					last_status = cur_status;
					cur_status->next = NULL;
				}
			}

			/* reload config */
			free(ctable);
			free_config_data(first);
			if(reload_config(argv[1], &first, &last, &firstg, &lastg)) {
				syslog(LOG_ERR, "reload config failed");
				exit(2);
			}
			init_config(first, last, proto, &ctable);

			/* restore connection statuses */
			for(cur_status = first_status; cur_status; cur_status = cur_status->next) {
				for(cur = first; cur; cur = cur->next) {
					if(strcmp(cur_status->name, cur->name) == 0) {
						((TARGET *)cur->data)->status = cur_status->status;
					}
				}
			}

			/* get rid of temporary statuses */
			cur_status = first_status;
			while(cur_status) {
				STATUS_DATA *tmp;

				tmp = cur_status->next;

				free(cur_status->name);
				free(cur_status);

				cur_status = tmp;
			}
		}

		for(cur = first; cur; cur = cur->next) {
			struct timeval current_time = {0, 0};

			if(start)
				while(wait_for_replies(ctable));

			if(gettimeofday(&current_time, NULL) == -1) {
				syslog(LOG_INFO, "gettimeofday failed \"%s\"", strerror(errno));
				sleep(1);
				continue;
			}

			t = cur->data;

			if(timeval_diff_cmp(&current_time, &last_sent_time, TIMEVAL_DIFF_CMP_LT, MIN_PERHOST_INTERVAL / 1000000L, MIN_PERHOST_INTERVAL % 1000000L)) continue;

			if(timeval_diff_cmp(&current_time, &(t->last_send_time), TIMEVAL_DIFF_CMP_LT, (cur->interval_ms * 1000) / 1000000L, (cur->interval_ms * 1000) % 1000000L)) continue;

			if(ping_send(cur)) {
				if(cfg.debug >= 9) syslog(LOG_INFO, "ping_send failed to %s", cur->name);
			}
			else {
				gettimeofday(&last_sent_time, NULL);
				start = 1;
			}
		}

		gettimeofday(&tv, NULL);
		if(timeval_diff_cmp(&tv, &last_decision, TIMEVAL_DIFF_CMP_GT, 1, 0)) { /* make decisions at 1s intervals */
			gettimeofday(&last_decision, NULL);
			decide(first);
			groups_decide(firstg);
		}
	} /* while cont */

	/* if we wrote pid file then close and remove it */
	if(lfp) {
		close(lfp);
		unlink(argv[2]);
	}

	free(ctable);
	free_config_data(first);
	free_config(&first, &last, &firstg, &lastg);
	closelog();

	return(0);
}

static void free_config_data(CONFIG *first) {
	CONFIG *cur;

	for(cur = first; cur; cur = cur->next) {
		TARGET *t;

		t = cur->data;
		close(t->sock);
		free(t);
	}
}

void decide(CONFIG *first) {
	struct timeval current_time = {0, 0};
	CONFIG *cur;
	int status_change = 0;

	gettimeofday(&current_time, NULL);

	for(cur = first; cur; cur = cur->next) {
		TARGET *t;
		int i, seq, ind;
		int timeout = 0, replied = 0, waiting = 0, reply_late = 0, used = 0, consecutive_waiting = 0, consecutive_missing = 0, consecutive_rcvd = 0;
		long rtt = 0;

		t = cur->data;

		/* check consecutive pkts */
		seq = t->seq % FOLLOWED_PKTS;

		for(i = (seq - 2); i > (seq - 2) - FOLLOWED_PKTS; i--) {
			ind = (i >= 0) ? i : i + FOLLOWED_PKTS;
			if(!t->sentpkts[ind].flags.used) break;

			if(t->sentpkts[ind].flags.waiting) consecutive_waiting++;
			else break;
		}

		for(i = (seq - 2); i > (seq - 2) - FOLLOWED_PKTS; i--) {
			ind = (i >= 0) ? i : i + FOLLOWED_PKTS;
			if(!t->sentpkts[ind].flags.used) break;

			if(t->sentpkts[ind].flags.timeout || t->sentpkts[ind].flags.waiting) consecutive_missing++;
			else break;
		}

		for(i = (seq - 2); i > (seq - 2) - FOLLOWED_PKTS; i--) {
			ind = (i >= 0) ? i : i + FOLLOWED_PKTS;
			if(!t->sentpkts[ind].flags.used) break;

			if(t->sentpkts[ind].flags.replied && !t->sentpkts[ind].flags.timeout) consecutive_rcvd++;
			else break;
		}

		/* count pkt states */
		for(i = 0; i < FOLLOWED_PKTS; i++) {
			if(!t->sentpkts[i].flags.used) continue;

			if(timeval_diff_cmp(&current_time, &t->sentpkts[i].sent_time, TIMEVAL_DIFF_CMP_GT, (cur->timeout_ms * 1000) / 1000000L, (cur->timeout_ms * 1000) % 1000000L) && t->sentpkts[i].flags.waiting) {
				t->sentpkts[i].flags.timeout = 1;
			}

			if(t->sentpkts[i].flags.replied && t->sentpkts[i].flags.timeout) reply_late++;

			if(t->sentpkts[i].flags.replied) {
				replied++;
				rtt += t->sentpkts[i].rtt; /* count avg rtt in usec from replied pkts rtt which is in usec */
			}
			if(t->sentpkts[i].flags.timeout) timeout++;
			if(t->sentpkts[i].flags.waiting) waiting++;
			used++;
		}

		/* make decision */
		if(cfg.debug >= 9) syslog(LOG_INFO, "name = %s, replied = %d, waiting = %d, timeout = %d, late reply = %d, cons rcvd = %d, cons wait = %d, cons miss = %d, avg_rtt = %.3f, seq = %d", cur->name, replied, waiting, timeout, reply_late, consecutive_rcvd, consecutive_waiting, consecutive_missing, (rtt / (replied ? replied : 1)) / 1000.0, t->seq);

		if(t->status) {
			if(timeout >= cur->max_packet_loss || consecutive_missing >= cur->max_successive_pkts_lost) {
				/* change to down state */
				if(cfg.debug >= 8) syslog(LOG_INFO, "link %s down event", cur->name);
				if(event_script_check(cur->eventscript))
					forkexec("%s %s %s %s %s %s %d %d %d %d %d %d %d %d",
						 cur->eventscript,
						 "down",
						 cur->name,
						 cur->checkip,
						 (cur->device && *cur->device) ? cur->device : "NA",
						 (cur->warn_email && *cur->warn_email) ? cur->warn_email : "NA",
						 replied,
						 waiting,
						 timeout,
						 reply_late,
						 consecutive_rcvd,
						 consecutive_waiting,
						 consecutive_missing,
						 rtt / (replied ? replied : 1));

				status_change = 1;
				t->status = 0;
			}
		}
		else {
			if(timeout <= cur->min_packet_loss && consecutive_rcvd >= cur->min_successive_pkts_rcvd) {
				/* change to up state */
				if(cfg.debug >= 8) syslog(LOG_INFO, "link %s up event", cur->name);
				if(event_script_check(cur->eventscript))
					forkexec("%s %s %s %s %s %s %d %d %d %d %d %d %d %d",
						 cur->eventscript,
						 "up",
						 cur->name,
						 cur->checkip,
						 (cur->device && *cur->device) ? cur->device : "NA",
						 (cur->warn_email && *cur->warn_email) ? cur->warn_email : "NA",
						 replied,
						 waiting,
						 timeout,
						 reply_late,
						 consecutive_rcvd,
						 consecutive_waiting,
						 consecutive_missing,
						 rtt / (replied ? replied : 1));

				status_change = 1;
				t->status = 1;
			}
		}

		if(dump || status_change) { /* dump is controlled by SIGUSR1 and then we should show all statuses anyway */
			char buf[FOLLOWED_PKTS + 100]; /* 100 should be enough for the comments and such, but I don't care to count */

			syslog(LOG_INFO, "name = %s, replied = %d, waiting = %d, timeout = %d, late reply = %d, cons rcvd = %d, cons wait = %d, cons miss = %d, avg_rtt = %.3f, seq = %d\n", cur->name, replied, waiting, timeout, reply_late, consecutive_rcvd, consecutive_waiting, consecutive_missing, (rtt / (replied ? replied : 1)) / 1000.0, t->seq);

			sprintf(buf, "seq        ");
			for(i = 0; i < FOLLOWED_PKTS; i++) {
				if(i == seq) strcat(buf, "*");
				else strcat(buf, " ");
			}
			syslog(LOG_INFO, "%s", buf);

			sprintf(buf, "used       ");
			for(i = 0; i < FOLLOWED_PKTS; i++) {
				sprintf(buf + strlen(buf), "%d", t->sentpkts[i].flags.used);
			}
			syslog(LOG_INFO, "%s", buf);

			sprintf(buf, "wait       ");
			for(i = 0; i < FOLLOWED_PKTS; i++) {
				sprintf(buf + strlen(buf), "%d", t->sentpkts[i].flags.waiting);
			}
			syslog(LOG_INFO, "%s", buf);

			sprintf(buf, "replied    ");
			for(i = 0; i < FOLLOWED_PKTS; i++) {
				sprintf(buf + strlen(buf), "%d", t->sentpkts[i].flags.replied);
			}
			syslog(LOG_INFO, "%s", buf);

			sprintf(buf, "timeout    ");
			for(i = 0; i < FOLLOWED_PKTS; i++) {
				sprintf(buf + strlen(buf), "%d", t->sentpkts[i].flags.timeout);
			}
			syslog(LOG_INFO, "%s", buf);

			sprintf(buf, "error      ");
			for(i = 0; i < FOLLOWED_PKTS; i++) {
				sprintf(buf + strlen(buf), "%d", t->sentpkts[i].flags.error);
			}
			syslog(LOG_INFO, "%s", buf);

			status_change = 0;
		}

	}
	if(dump) dump = 0; /* if we just dumped then don't dump next time. flags don't change that frequently */
}

void groups_decide(GROUPS *firstg){
	GROUPS *curg;
	GROUP_MEMBERS *curgm;
	TARGET *t;
	int status;


	curg = firstg;
	while(curg) {
		status = curg->logic;

		curgm = curg->fgm;
		while(curgm) {
			if(!curgm->cfg_ptr) break;

			t = curgm->cfg_ptr->data;

			if(!curg->logic) {
				status |= t->status;
			} else {
				status &= t->status;
			}

			curgm = curgm->next;
		}
		if(status != curg->status) {
			curg->status = status;
			if(curg->status) {
				/* group up event */
				if(cfg.debug >= 8) syslog(LOG_INFO, "group %s up event", curg->name);
				if(event_script_check(curg->eventscript))
					forkexec("%s %s %s %s %s %s %d %d %d %d %d %d %d %d",
						 curg->eventscript,
						 "up",
						 curg->name,
						 "NA",
						 "NA",
						 (curg->warn_email && *curg->warn_email) ? curg->warn_email : "NA",
						 0,
						 0,
						 0,
						 0,
						 0,
						 0,
						 0,
						 0);
			} else {
				/* group down event */
				if(cfg.debug >= 8) syslog(LOG_INFO, "group %s down event", curg->name);
				if(event_script_check(curg->eventscript))
					forkexec("%s %s %s %s %s %s %d %d %d %d %d %d %d %d",
						 curg->eventscript,
						 "down",
						 curg->name,
						 "NA",
						 "NA",
						 (curg->warn_email && *curg->warn_email) ? curg->warn_email : "NA",
						 0,
						 0,
						 0,
						 0,
						 0,
						 0,
						 0,
						 0);
			}
		}
		curg = curg->next;
	}
}

int wait_for_replies(CONFIG **ctable) {
	struct ip *ip;
	int hlen = 0;
	struct icmp *icp;
	PING_DATA *pdp;
	int this_count;
	struct timeval sent_time = {0, 0};
	struct timeval current_time = {0, 0};
	long time_diff;
	char buf[BUFSIZ];
	union {
		struct sockaddr_in saddr;
		struct sockaddr_ll FROM;
	} from_addr;
	unsigned int slen;
	int result;
	TARGET *t;
	int seq;
	CONFIG *arp;

	slen = sizeof(from_addr);
	result = ping_rcv(ctable[0], buf, BUFSIZ, (struct sockaddr_in *)&from_addr, &slen, DEFAULT_SELECT_WAIT, &arp);

	if(result < 0) {
		return(0);
	}

	gettimeofday(&current_time, NULL);

	if(arp) {
		struct sockaddr_ll *FROM = &from_addr.FROM;
		TARGET *t = arp->data;
		struct arphdr *ah = (struct arphdr*)buf;
		unsigned char *p = (unsigned char *)(ah+1);
		struct in_addr src_ip, dst_ip;
		int ind;

		/* Filter out wild packets */
		if(FROM->sll_pkttype != PACKET_HOST &&
		   FROM->sll_pkttype != PACKET_BROADCAST &&
		   FROM->sll_pkttype != PACKET_MULTICAST)
			return(1);

		/* Only these types are recognised */
#if 0
		if(ah->ar_op != htons(ARPOP_REQUEST) &&
		   ah->ar_op != htons(ARPOP_REPLY))
			return(1);
#else
		if(ah->ar_op != htons(ARPOP_REPLY))
			return(1);
#endif

		/* ARPHRD check and this darned FDDI hack here :-( */
		if(ah->ar_hrd != htons(FROM->sll_hatype) &&
		   (FROM->sll_hatype != ARPHRD_FDDI || ah->ar_hrd != htons(ARPHRD_ETHER)))
			return(1);

		/* Protocol must be IP. */
		if(ah->ar_pro != htons(ETH_P_IP))
			return(1);
		if(ah->ar_pln != 4)
			return(1);
		if(ah->ar_hln != t->me.sll_halen)
			return(1);

#if defined(DEBUG)
		for(ind = 0; ind < result; ind ++) {
			fprintf(stderr, "%02x", (unsigned char)buf[ind]);
			if(!((ind + 1) % 2)) fprintf(stderr, " ");
			if(!((ind + 1) % 32)) fprintf(stderr, "\n");
		}
		fprintf(stderr, "\n");
		for(ind = 0; ind < result; ind ++) {
			fprintf(stderr, "%3u", (unsigned char)buf[ind]);
			if(!((ind + 1) % 32)) fprintf(stderr, "\n");
			else fprintf(stderr, ",");
		}
		fprintf(stderr, "\n");
#endif

		if(result < sizeof(*ah) + 2*(4 + ah->ar_hln))
			return(1);

		memcpy(&src_ip, p+ah->ar_hln, 4);
		memcpy(&dst_ip, p+ah->ar_hln+4+ah->ar_hln, 4);

		if(src_ip.s_addr != t->dst.s_addr)
			return(1);
		if(t->src.s_addr != dst_ip.s_addr)
			return(1);
		if(memcmp(p+ah->ar_hln+4, &t->me.sll_addr, ah->ar_hln))
			return(1);

		/* update packet log here */
		/* there are no sequence numbers in arp replies so just mark seq - 1 replied */
		ind = ((t->seq - 1) >= 0 ? (t->seq - 1) : (FOLLOWED_PKTS + (t->seq - 1))) % FOLLOWED_PKTS;
		t->sentpkts[ind].flags.replied = 1;
		t->sentpkts[ind].flags.waiting = 0;
		t->sentpkts[ind].replied_time = current_time;
		t->sentpkts[ind].rtt = timeval_diff(&current_time, &t->sentpkts[ind].sent_time);

		return(1);
	}

	/* ping reply */
	ip = (struct ip *)buf;
	hlen = ip->ip_hl << 2;

	icp = (struct icmp *)(buf + hlen);

	if(icp->icmp_type == ICMP_ECHO) {
		return(1);
	}

	if(icp->icmp_type != ICMP_ECHOREPLY) {
		if(cfg.debug >= 9) syslog(LOG_INFO, "got odd reply from %s, icmp_type = %s, icmp_code = %s", inet_ntoa(from_addr.saddr.sin_addr), stricmptype(icp->icmp_type), icp->icmp_code == ICMP_UNREACH ? stricmpcode(icp->icmp_code) : "none");
		/* handle here this odd pkt */
		handle_odd_icmp(icp, result - hlen, &from_addr.saddr);
		return(1);
	}

	if(icp->icmp_id != ident) {
		/* fprintf(stderr, "icmp_id = %d funny, got reply from %s to something else ...\n", icp->icmp_id, inet_ntoa(saddr.sin_addr)); */
		return(1);
	}

	if(result < sizeof(struct icmp) + sizeof(PING_DATA)) {
		/* fprintf(stderr, "too short ping reply\n"); */
		return(1);
	}

	pdp = (PING_DATA *)(buf + hlen + sizeof(struct icmp));

	this_count = pdp->ping_count;
	sent_time = pdp->ping_ts;
	time_diff = timeval_diff(&current_time, &sent_time);

	t = ctable[pdp->id]->data;
	seq = icp->icmp_seq % FOLLOWED_PKTS;
	if(t->sentpkts[seq].seq == icp->icmp_seq) {
		t->sentpkts[seq].flags.replied = 1;
		t->sentpkts[seq].flags.waiting = 0;
		t->sentpkts[seq].replied_time = current_time;
		t->sentpkts[seq].rtt = timeval_diff(&current_time, &t->sentpkts[seq].sent_time);
	}
	else
		if(cfg.debug >= 9) syslog(LOG_INFO, "sentpkts seq != icmp_seq");

	if(cfg.debug >= 9) syslog(LOG_INFO, "received seq = %d from %s, id = %d, num_sent = %d, target id = %u, time_diff = %ld", icp->icmp_seq, inet_ntoa(from_addr.saddr.sin_addr), icp->icmp_id, this_count, pdp->id, time_diff);

	return(1);
}

int ping_rcv(CONFIG *first, char *buf, int len, struct sockaddr_in *saddr, unsigned int *slen, long usec, CONFIG **arp) {
	int nfound, n;
	fd_set readset;
	struct timeval to = {0, 0};
	int max;
	CONFIG *cur;
	TARGET *t;

	FD_ZERO(&readset);
	max = 0;

	for(cur = first; cur; cur = cur->next) {
		t = cur->data;
		if(t->sock > max) max = t->sock;
		FD_SET(t->sock, &readset);
	}

	to.tv_sec = usec / 1000000;
	to.tv_usec = (usec - (to.tv_sec * 1000000));

#if defined(DEBUG)
	printf("to.tv_sec = %ld, to.tv_usec = %ld\n", to.tv_sec, to.tv_usec);
#endif

	nfound = select(max + 1, &readset, NULL, NULL, &to);

	if(nfound < 0) {
		if(errno != EINTR) syslog(LOG_INFO, "select failed \"%s\"", strerror(errno));
		return(0);
	}

	if(nfound == 0) return(-1);

	for(cur = first; cur; cur = cur->next) {
		t = cur->data;
		if(FD_ISSET(t->sock, &readset)) {
			if(!cur->check_arp) {
				*arp = (CONFIG *)NULL;
			} else {
				*arp = cur;
			}

			n = recvfrom(t->sock, buf, len, 0, (struct sockaddr *)saddr, slen);

			if(n < 0) {
				syslog(LOG_INFO, "arping recvfrom failed with %s \"%s\"\n", cur->name, strerror(errno));
				return(0);
			}

			return(n);
		}
	}
	return(0);
}

int ping_send(CONFIG *cur) {
	char buf[BUFSIZ];
	struct icmp *icp;
	PING_DATA *pdp;
	TARGET *t;
	int n;
	int ping_pkt_size;

	t = cur->data;

	gettimeofday(&t->last_send_time, NULL);

	if(cur->check_arp) {
		int err;
		unsigned char buf[256];
		struct arphdr *ah = (struct arphdr*)buf;
		unsigned char *p = (unsigned char *)(ah+1);

		ah->ar_hrd = htons(t->me.sll_hatype);
		if(ah->ar_hrd == htons(ARPHRD_FDDI))
			ah->ar_hrd = htons(ARPHRD_ETHER);
		ah->ar_pro = htons(ETH_P_IP);
		ah->ar_hln = t->me.sll_halen;
		ah->ar_pln = 4;
		ah->ar_op = htons(ARPOP_REQUEST);

		memcpy(p, &t->me.sll_addr, ah->ar_hln);
		p += t->me.sll_halen;

		memcpy(p, &t->src, 4);
		p += 4;

		memcpy(p, &t->he.sll_addr, ah->ar_hln);
		p += ah->ar_hln;

		memcpy(p, &t->dst, 4);
		p += 4;

#if defined(DEBUG)
		{
			int ind;
			for(ind = 0; ind < p - buf; ind ++) {
				fprintf(stderr, "%02x", (unsigned char)buf[ind]);
				if(!((ind + 1) % 2)) fprintf(stderr, " ");
				if(!((ind + 1) % 32)) fprintf(stderr, "\n");
			}
			fprintf(stderr, "\n");
			for(ind = 0; ind < p - buf; ind ++) {
				fprintf(stderr, "%3u", (unsigned char)buf[ind]);
				if(!((ind + 1) % 32)) fprintf(stderr, "\n");
				else fprintf(stderr, ",");
			}
			fprintf(stderr, "\n");
		}
#endif

		err = sendto(t->sock, buf, p - buf, 0, (struct sockaddr*)&t->he, sizeof(t->he));

		{ /* we don't care what the error was just advance with seq */
			int seq;

			seq = t->seq % FOLLOWED_PKTS;
			t->sentpkts[seq].seq = t->seq;
			t->sentpkts[seq].sent_time = t->last_send_time;
			t->sentpkts[seq].flags.replied = 0;
			t->sentpkts[seq].flags.timeout = 0;
			t->sentpkts[seq].flags.waiting = 1;
			t->sentpkts[seq].flags.used = 1;
			t->sentpkts[seq].flags.error = (err == -1) ? 1 : 0;

			t->seq = (t->seq + 1) % SEQ_LIMITER; /* limit seq so that consecutive missing and received pkt counting doesn't get confused when seq "overflows" */
			t->num_sent++;

		}
		if(err == (p - buf)) {
			return(0);
		}
		return(err);
	}

	/* send a ping packet */
	ping_pkt_size = sizeof(struct icmp) + sizeof(PING_DATA);

	memset(buf, 0, ping_pkt_size);

	icp = (struct icmp *)buf;

	icp->icmp_type = ICMP_ECHO;
	icp->icmp_code = 0;
	icp->icmp_cksum = 0;
	icp->icmp_seq = t->seq;
	icp->icmp_id = ident;

	pdp = (PING_DATA *)(buf + sizeof(struct icmp));
	pdp->ping_count = t->num_sent;
	pdp->ping_ts = t->last_send_time;
	pdp->id = t->id;

	icp->icmp_cksum = in_cksum((u_short *)icp, ping_pkt_size);

	n = sendto(t->sock, buf, ping_pkt_size, 0, (struct sockaddr *)&t->dst_addr, sizeof(struct sockaddr));

	if(n < 0) {
		if(errno == ENODEV) {
			if(cfg.debug >= 9) syslog(LOG_ERR, "connection %s no such device %s \"%s\"", cur->name, cur->device, strerror(errno));
			/* exit(2); */ /* commented out. handle this situation like the packet had been sent. see below.  */
			if(cfg.reopen_on_enodev) {
				/* try to reopen device */
				int sock;

				if((sock = reopen(cur)) != -1) {
					close(t->sock);
					t->sock = sock;
				}
			}
		}
		else
			if(cfg.debug >= 9) syslog(LOG_ERR, "ping sendto failed to %s on %s reason \"%s\"", cur->name, cur->device, strerror(errno));
	}

	{ /* we don't care what the error was just advance with seq */
		int seq;

		seq = t->seq % FOLLOWED_PKTS;
#if defined(DEBUG)
		fprintf(stderr, "ping_send seq = %ld to %s, num_sent = %ld, %ld, pkt_size = %d\n", t->seq, inet_ntoa(t->saddr.sin_addr), t->num_sent, pdp->ping_count, ping_pkt_size);
#endif
		t->sentpkts[seq].seq = t->seq;
		t->sentpkts[seq].sent_time = t->last_send_time;
		t->sentpkts[seq].flags.replied = 0;
		t->sentpkts[seq].flags.timeout = 0;
		t->sentpkts[seq].flags.waiting = 1;
		t->sentpkts[seq].flags.used = 1;
		t->sentpkts[seq].flags.error = (n < 1) ? 1 : 0;

		t->seq = (t->seq + 1) % SEQ_LIMITER; /* limit seq so that consecutive missing and received pkt counting doesn't get confused when seq "overflows" */
		t->num_sent++;
	}

	if(n == ping_pkt_size) {
		return(0);
	}
	return(n);
}

int handle_odd_icmp(struct icmp *icp, int icpsize, struct sockaddr_in *saddr) {

	return(0);
}

int in_cksum(u_short *p, int n) {
	register u_short answer;
	register long sum = 0;
	u_short odd_byte = 0;

	while(n > 1) {
		sum += *p++;
		n -= 2;
	}

	/* mop up an odd byte, if necessary */
	if(n == 1){
		*(u_char*)(&odd_byte) = *(u_char*)p;
		sum += odd_byte;
	}

	sum = (sum >> 16) + (sum & 0xffff);   /* add hi 16 to low 16 */
	sum += (sum >> 16);                   /* add carry */
	answer = ~sum;                        /* ones-complement, truncate*/

	return(answer);
}

int timeval_diff_cmp(struct timeval *a, struct timeval *b, int operation, time_t sec, suseconds_t usec) {
	time_t diff_sec;
	suseconds_t diff_usec;

	diff_sec = a->tv_sec - b->tv_sec;
	diff_usec = a->tv_usec - b->tv_usec;

	if(diff_usec < 0) {
		diff_sec--;
		diff_usec = 1000000L - diff_usec;
	}

	switch(operation) {
	case TIMEVAL_DIFF_CMP_GT:
		if(diff_sec > sec) return(TRUE);
		if(diff_sec == sec && diff_usec > usec) return(TRUE);
		return(FALSE);
		break;
	case TIMEVAL_DIFF_CMP_LT:
		if(diff_sec < sec) return(TRUE);
		if(diff_sec == sec && diff_usec < usec) return(TRUE);
		return(FALSE);
		break;
	default:
		syslog(LOG_ERR, "warning: unknown timeval_diff_cmp operation requested %d", operation);
		return(FALSE);
	}
}

long timeval_diff(struct timeval *a, struct timeval *b)
{
	return( ((a->tv_sec - b->tv_sec) * 1000000L) + (a->tv_usec - b->tv_usec) );
}

void usage(void) {
	printf("%s <config.file>\n", prog); /* ... */
}

void forkexec(char *fmt, ...) {
	pid_t pid;
	va_list vl;
	char buf[BUFSIZ];
	char *argv[64];
	int exitval;
	char *envp[64];
	char envbuf[BUFSIZ];
#if defined(DEBUG)
	int i;
#endif

	/* fill in buf */
	va_start(vl, fmt);
	vsprintf(buf, fmt, vl);
	va_end(vl);

	/* fill in envbuf */
	sprintf(envbuf, "LANG=%s PATH=%s TERM=%s", getenv("LANG"), getenv("PATH"), getenv("TERM"));

	pid = fork();

	if(pid == -1) {
		syslog(LOG_ERR, "forkexec: fork() failed \"%s\"", strerror(errno));
		return;
	}

	if(pid) {
		/* parent */
		wait(&exitval);
		if(cfg.debug >= 9 && exitval) syslog(LOG_ERR, "script exited with non null exit value %d", exitval);
		return;
	}

	/* child */
	closelog(); /* openlog doesn't return a fd to set close on exec */

	/* make argv pointers */
	makeplist(argv, buf);

	/* make envp pointers */
	makeplist(envp, envbuf);

#if defined(DEBUG)
	for(i = 0; argv[i]; i++) {
		fprintf(stderr, "argv[%d] = \"%s\"\n", i, argv[i]);
	}
	for(i = 0; envp[i]; i++) {
		fprintf(stderr, "envp[%d] = \"%s\"\n", i, envp[i]);
	}
#endif

	execve(argv[0], argv, envp);
	exit(1); /* exec failed ... */
}

void makeplist(char *plist[], char *buf)
{
	int i,ap;

	for(i = ap = 0; buf[i]; ap++) {
		plist[ap] = buf + i;
		for(; buf[i] && buf[i] != ' '; i++);
		if(buf[i]) buf[i++] = '\0';
	}
	plist[ap] = NULL;
}

int event_script_check(const char *path)
{
	struct stat statbuf;

	if(!path) {
		if(cfg.debug >= 9) syslog(LOG_ERR, "NULL pointer event script");
		return(0);
	}

	if(!*path) {
		if(cfg.debug >= 9) syslog(LOG_ERR, "null string event script");
		return(0);
	}

	/* check that the script is owner executable */
	if(stat(path, &statbuf) == -1) {
		syslog(LOG_ERR, "failed to stat event script \"%s\" reason \"%s\"", path, strerror(errno));
		return(0);
	}

	if((statbuf.st_mode & S_IXUSR) == 0) {
		syslog(LOG_ERR, "event script \"%s\" is not executable by owner, please check permissions", path);
		return(0);
	}

	return(1);
}

void signal_handler(int signo) {
	int errno_save;

	errno_save = errno;

	switch(signo) {
	case SIGINT:
		cont = 0;
		break;
	case SIGUSR1:
		dump = 1;
		break;
	case SIGHUP:
		reload_cfg = 1;
		break;
	}

	errno = errno_save;
}

static int reopen(CONFIG *cur)
{
	int sock;
	struct protoent *proto;

	if((proto = getprotobyname("icmp")) == NULL) {
		syslog(LOG_ERR, "no icmp proto found on reopen");
		return(-1);
	}

	sock = socket(AF_INET, SOCK_RAW, proto->p_proto);

	if(sock < 0) {
		syslog(LOG_ERR, "could not reopen socket for ping target \"%s\" reason \"%s\"\n", cur->name, strerror(errno));
		return(-1);
	}

	if(fcntl(sock, F_SETFD, FD_CLOEXEC) == -1) {
		syslog(LOG_ERR, "failed to set close on exec on socket %s after reopen reason \"%s\"", cur->name, strerror(errno));
		close(sock);
		return(-1);
	}

	if(cur->ttl) {
		int ittl = cur->ttl;
		if(setsockopt(sock, IPPROTO_IP, IP_MULTICAST_TTL,
			      &cur->ttl, 1) == -1) {
			syslog(LOG_ERR, "can't set multicast time-to-live on reopen \"%s\"", strerror(errno));
			close(sock);
			return(-1);
		}
		if(setsockopt(sock, IPPROTO_IP, IP_TTL,
			      &ittl, sizeof(ittl)) == -1) {
			syslog(LOG_ERR, "can't set unicast time-to-live on reopen \"%s\"", strerror(errno));
			close(sock);
			return(-1);
		}
	}

	if(cur->device && *cur->device) {
		if(setsockopt(sock, SOL_SOCKET, SO_BINDTODEVICE, cur->device, strlen(cur->device) + 1) == -1)
			syslog(LOG_INFO, "WARNING: on reopen ping interface \"%s\" is ignored \"%s\"", cur->device, strerror(errno));
	}

	return(sock);
}

static void init_config(CONFIG *first, CONFIG *last, struct protoent *proto, CONFIG ***ctable)
{
	int i;
	int num_hosts = 0;
	CONFIG *cur;
	TARGET *t = NULL;

	/* initialize config->data */
	for(cur = first, num_hosts = 0; cur; cur = cur->next, num_hosts++) {
		u_int ipaddress;

		if((t = malloc(sizeof(TARGET))) == NULL) {
			syslog(LOG_ERR, "main: initializing targets failed to malloc");
			exit(1);
		}
		memset(t, 0, sizeof(TARGET));

		ipaddress = inet_addr(cur->checkip);
		t->seq = 0;
		t->last_send_time.tv_sec = 0;
		t->last_send_time.tv_usec = 0;
		t->num_sent = 0;
		t->dst_addr.sin_family = AF_INET;
		t->dst_addr.sin_addr = *((struct in_addr *)&ipaddress);

		t->id = num_hosts;

		/* get initial connection state assumption from config */
		t->status = cur->status;

		if(cur->check_arp) {
			int ifindex = 0;

			t->sock = socket(PF_PACKET, SOCK_DGRAM, 0);
			if(t->sock < 0) {
				syslog(LOG_ERR, "could not open socket for %s arp ping \"%s\"", cur->name, strerror(errno));
				exit(1);
			}
			if(fcntl(t->sock, F_SETFD, FD_CLOEXEC) == -1) {
				syslog(LOG_ERR, "failed to set close on exec on socket %s reason \"%s\"", cur->name, strerror(errno));
			}

			if(cur->device && *cur->device) {
				struct ifreq ifr;

				memset(&ifr, 0, sizeof(ifr));
				strncpy(ifr.ifr_name, cur->device, IFNAMSIZ-1);
				if(ioctl(t->sock, SIOCGIFINDEX, &ifr) < 0) {
					syslog(LOG_ERR, "unknown iface \"%s\"", cur->device);
					exit(2);
				}
				ifindex = ifr.ifr_ifindex;

				if(ioctl(t->sock, SIOCGIFFLAGS, (char*)&ifr)) {
					syslog(LOG_ERR, "ioctl(SIOCGIFFLAGS) \"%s\"", strerror(errno));
					exit(2);
				}
				if(!(ifr.ifr_flags&IFF_UP)) {
					syslog(LOG_ERR, "Interface \"%s\" is down", cur->device);
					exit(2);
				}
				if(ifr.ifr_flags&(IFF_NOARP|IFF_LOOPBACK)) {
					syslog(LOG_ERR, "Interface \"%s\" is not ARPable", cur->device);
					exit(2);
				}
			}

			if(inet_aton(cur->checkip, &t->dst) != 1) {
				struct hostent *hp;
				hp = gethostbyname2(cur->checkip, AF_INET);
				if(!hp) {
					syslog(LOG_ERR, "unknown host %s\n", cur->checkip);
					exit(2);
				}
				memcpy(&t->dst, hp->h_addr, 4);
			}

			if(cur->sourceip && *cur->sourceip)
				if(inet_aton(cur->sourceip, &t->src) != 1) {
					syslog(LOG_ERR, "invalid source %s\n", cur->sourceip);
					exit(2);
				}

#if 0 /* there is no unsolicited mode for lsm */
			if(t->src.s_addr == 0)
				t->src = t->dst;
#endif

			if(1 || t->src.s_addr) { /* there is no dad mode either. do this always */
				struct sockaddr_in saddr;
				int probe_fd = socket(AF_INET, SOCK_DGRAM, 0);

				if(probe_fd < 0) {
					syslog(LOG_ERR, "socket \"%s\"", strerror(errno));
					exit(2);
				}
				if(fcntl(t->sock, F_SETFD, FD_CLOEXEC) == -1) {
					syslog(LOG_ERR, "failed to set close on exec on probe socket reason \"%s\"", strerror(errno));
				}

				if(cur->device && *cur->device) {
					if(setsockopt(probe_fd, SOL_SOCKET, SO_BINDTODEVICE, cur->device, strlen(cur->device) + 1) == -1)
						syslog(LOG_INFO, "WARNING: arping interface \"%s\" is ignored \"%s\"", cur->device, strerror(errno));
				}
				memset(&saddr, 0, sizeof(saddr));
				saddr.sin_family = AF_INET;
				if(t->src.s_addr) {
					saddr.sin_addr = t->src;
					if(bind(probe_fd, (struct sockaddr*)&saddr, sizeof(saddr)) == -1) {
						syslog(LOG_ERR, "bind \"%s\"", strerror(errno));
						exit(2);
					}
				}
				else {
					int on = 1;
					int alen = sizeof(saddr);

					saddr.sin_port = htons(1025);
					saddr.sin_addr = t->dst;

					if(setsockopt(probe_fd, SOL_SOCKET, SO_DONTROUTE, (char*)&on, sizeof(on)) == -1)
						syslog(LOG_INFO, "WARNING: setsockopt(SO_DONTROUTE) \"%s\"", strerror(errno));
					if(connect(probe_fd, (struct sockaddr*)&saddr, sizeof(saddr)) == -1) {
						syslog(LOG_ERR, "connect \"%s\"", strerror(errno));
						exit(2);
					}
					if(getsockname(probe_fd, (struct sockaddr*)&saddr, (socklen_t*)&alen) == -1) {
						syslog(LOG_ERR, "getsockname \"%s\"", strerror(errno));
						exit(2);
					}
					t->src = saddr.sin_addr;
				}
				close(probe_fd);
			}

			t->me.sll_family = AF_PACKET;
			t->me.sll_ifindex = ifindex;
			t->me.sll_protocol = htons(ETH_P_ARP);
			if(bind(t->sock, (struct sockaddr*)&t->me, sizeof(t->me)) == -1) {
				syslog(LOG_ERR, "bind \"%s\"", strerror(errno));
				exit(2);
			}

			if(1) {
				int alen = sizeof(t->me);
				if(getsockname(t->sock, (struct sockaddr*)&t->me, (socklen_t*)&alen) == -1) {
					syslog(LOG_ERR, "getsockname \"%s\"", strerror(errno));
					exit(2);
				}
			}
			if(t->me.sll_halen == 0) {
				syslog(LOG_ERR, "Interface \"%s\" is not ARPable (no ll address)", cur->device);
				exit(2);
			}

			t->he = t->me;
			memset(t->he.sll_addr, -1, min(t->he.sll_halen, sizeof t->he.sll_addr));

#if 0
			printf("ARPING %s ", inet_ntoa(t->dst));
			printf("from %s %s\n",  inet_ntoa(t->src), cur->device ? : "");
#endif

			if(!t->src.s_addr) {
				syslog(LOG_ERR, "no source address for %s", cur->name);
				exit(2);
			}
			if(cur->ttl) {
				int ittl = cur->ttl;
				if(setsockopt(t->sock, IPPROTO_IP, IP_MULTICAST_TTL,
					      &cur->ttl, 1) == -1) {
					syslog(LOG_ERR, "can't set multicast time-to-live \"%s\"", strerror(errno));
					exit(2);
				}
				if(setsockopt(t->sock, IPPROTO_IP, IP_TTL,
					      &ittl, sizeof(ittl)) == -1) {
					syslog(LOG_ERR, "can't set unicast time-to-live \"%s\"", strerror(errno));
					exit(2);
				}
			}
		} else {

			t->sock = socket(AF_INET, SOCK_RAW, proto->p_proto);

			if(t->sock < 0) {
				syslog(LOG_ERR, "could not open socket for ping target \"%s\" reason \"%s\"\n", cur->name, strerror(errno));
				exit(1);
			}
			if(fcntl(t->sock, F_SETFD, FD_CLOEXEC) == -1) {
				syslog(LOG_ERR, "failed to set close on exec on socket %s reason \"%s\"", cur->name, strerror(errno));
			}

			if(cur->ttl) {
				int ittl = cur->ttl;
				if(setsockopt(t->sock, IPPROTO_IP, IP_MULTICAST_TTL,
					      &cur->ttl, 1) == -1) {
					syslog(LOG_ERR, "can't set multicast time-to-live \"%s\"", strerror(errno));
					exit(2);
				}
				if(setsockopt(t->sock, IPPROTO_IP, IP_TTL,
					      &ittl, sizeof(ittl)) == -1) {
					syslog(LOG_ERR, "can't set unicast time-to-live \"%s\"", strerror(errno));
					exit(2);
				}
			}

			if(cur->device && *cur->device) {
				if(setsockopt(t->sock, SOL_SOCKET, SO_BINDTODEVICE, cur->device, strlen(cur->device) + 1) == -1)
					syslog(LOG_INFO, "WARNING: ping interface \"%s\" is ignored \"%s\"", cur->device, strerror(errno));
			}
		}
		cur->data = t;
	}

	if(((*ctable) = (CONFIG **)malloc(sizeof(CONFIG *) * num_hosts)) == NULL) {
		syslog(LOG_ERR, "main: can't malloc for ctable");
		exit(1);
	}

	/* create pointer table */
	for(cur = first, i = 0; cur; cur = cur->next, i++) {
		(*ctable)[i] = cur;
	}

}

/* EOF */
